from __future__ import (absolute_import, division, print_function)

import os


def resource_file(filepath):
    return os.path.join(test_directory(), 'resources', filepath)


def test_directory():
    """Helper function to return path to the tests directory"""
    return os.path.dirname(__file__)



from __future__ import (absolute_import, division, print_function)

import pytz
from owslib.etree import ElementType, etree

from owslib.waterml.wml11 import WaterML_1_1
from shapely.geometry import Point as sPoint
from paegan.cdm.dsg.features.base.point import Point
from paegan.cdm.dsg.features.station import Station as Station
from paegan.cdm.dsg.collections.station_collection import StationCollection
from paegan.cdm.dsg.member import Member


class WaterML11ToPaegan(object):
    def __init__(self, waterml_data):

        if isinstance(waterml_data, ElementType):
            self._root = waterml_data
        else:
            self._root = etree.fromstring(waterml_data.encode())

        response = WaterML_1_1(self._root).response

        stations = []
        station_lookup = []

        for timeseries in response.time_series:
            station_code = sorted(timeseries.source_info.site_codes)[0]
            # Create station if we have not seen it
            if station_code not in station_lookup:
                s = Station()
                s.uid = station_code

                info = timeseries.source_info
                s.name = info.site_name
                s.set_property("station_type", info.site_types)
                s.set_property("huc", info.site_properties.get("hucCd"))
                s.set_property("county", info.site_properties.get("countyCd"))
                s.set_property("state", info.site_properties.get("stateCd"))

                # Now set the station's location
                vertical = info.elevation
                if vertical is None:
                    vertical = 0

                try:
                    location = info.location.geo_coords[0]
                    srs = info.location.srs[0]
                except:
                    print("Could not find a location for {}... skipping station".format(s.uid))
                    continue

                s.location = sPoint(float(location[0]), float(location[1]), vertical)
                s.set_property("horizontal_crs", srs)
                s.set_property("vertical_units", "m")
                s.set_property("vertical_crs", info.vertical_datum)
                s.set_property("location_description", info.location.notes)

                stations.append(s)
                station_lookup.append(s.uid)

            times = {}
            variable = timeseries.variable
            for variable_timeseries in timeseries.values:
                for r in variable_timeseries:
                    dt = r.date_time
                    if dt.tzinfo is None:
                        dt = dt.replace(tzinfo=pytz.utc)
                    dt = dt.astimezone(pytz.utc)

                    if dt not in times.keys():
                        times[dt] = []

                    times[dt].append(Member(value=r.value, unit=variable.unit.code, name=variable.variable_name, description=variable.variable_description, standard=variable.variable_code))

            station = stations[station_lookup.index(station_code)]
            for dts, members in times.items():
                p = Point()
                p.time = dts
                p.location = station.location
                p.members = members
                station.add_element(p)

        self.feature = StationCollection(elements=stations)



from __future__ import (absolute_import, division, print_function)
from six import text_type, string_types

from pyoos.collectors.ioos.swe_sos import IoosSweSos


class NdbcSos(IoosSweSos):
    def __init__(self, **kwargs):
        if kwargs.get("test", None) is True:
            kwargs["url"] = "http://sdftest.ndbc.noaa.gov/sos/server.php"
        else:
            kwargs["url"] = 'http://sdf.ndbc.noaa.gov/sos/server.php'

        super(NdbcSos, self).__init__(**kwargs)
        self._datum = None
        self._dataType = None

    def setup_params(self, **kwargs):
        params = super(NdbcSos, self).setup_params(**kwargs)

        if self.bbox is not None:
            params["featureofinterest"] = "BBOX:%s" % ",".join([text_type(x) for x in self.bbox])

        if self.features is None or len(self.features) < 1:
            params["offerings"] = ["urn:ioos:network:noaa.nws.ndbc:all"]
        elif len(self.features) > 1:
            # TODO: Send many requests, one for each station, rather than a network:all
            print("NDBC does not support filtering by > 1 station at a time... returning all stations.")
            params["offerings"] = ["urn:ioos:network:noaa.nws.ndbc:all"]
        elif len(self.features) == 1:
            params["offerings"] = ["urn:ioos:station:wmo:%s" % self.features[0]]

        if params.get("responseFormat", None) is None:
            params["responseFormat"] = 'text/csv'

        if self.variables is None or len(self.variables) < 1:
            raise ValueError("You must set a filter for at least one variable (observedProperty)")
        else:
            ops = ",".join(self.variables)
            if isinstance(ops, string_types):
                ops = [ops]
            params["observedProperties"] = ops

        return params

    def metadata(self, **kwargs):
        callback = lambda x: "urn:ioos:station:wmo:%s" % x
        return super(NdbcSos, self).metadata(feature_name_callback=callback, **kwargs)



from __future__ import (absolute_import, division, print_function)

def flatten_element(p):
    """
    Convenience function to return record-style time series representation
    from elements ('p') members in station element.
    member['standard'] is a standard_name parameter name, typically CF based.
    Ideally, member['value'] should already be floating point value,
    so it's ready to use.
    Useful with most pyoos collectors.
    """
    rd = {'time': p.time}
    for member in p.members:
        rd[member['standard']] = member['value']
    return rd



from __future__ import (absolute_import, division, print_function)

import unittest
from pyoos.collectors.collector import Collector
from datetime import datetime

from pytz import timezone
from pytz.reference import Eastern


class CollectorTest(unittest.TestCase):
    def test_times_return_utc(self):
        c = Collector()

        # Assumes timezones without a timezone are UTC
        c.filter(start=datetime(2000, 2, 1, 10), end=datetime(2000, 2, 3, 8))
        assert c._start_time == datetime(2000, 2, 1, 10, tzinfo=timezone('UTC'))
        assert c._end_time == datetime(2000, 2, 3, 8, tzinfo=timezone('UTC'))

        # Always returns UTC (5 hour difference here)
        c.filter(start=datetime(2000, 2, 1, 10, tzinfo=Eastern), end=datetime(2000, 2, 3, 8, tzinfo=Eastern))
        assert c._start_time == datetime(2000, 2, 1, 15, tzinfo=timezone('UTC'))
        assert c._end_time == datetime(2000, 2, 3, 13, tzinfo=timezone('UTC'))



from __future__ import (absolute_import, division, print_function)

from collections import defaultdict
from itertools import groupby
from dateutil.parser import parser
import time
import pytz

from paegan.cdm.dsg.member import Member
from paegan.cdm.dsg.features.station import Station
from paegan.cdm.dsg.collections.station_collection import StationCollection
from paegan.cdm.dsg.features.base.point import Point
from shapely.geometry import Point as sPoint
from numpy import nan as npNan


class HadsParser(object):

    def __init__(self):
        pass

    def parse(self, metadata, raw_data, var_filter, time_extents):

        self.parsed_metadata = self._parse_metadata(metadata)
        self.parsed_data     = self._parse_data(raw_data, var_filter, time_extents)
        self.feature         = self._build_station_collection(self.parsed_metadata, self.parsed_data)

        return self.feature

    def _build_station_collection(self, parsed_metadata, parsed_data):

        stations = []
        for station_code, station_metadata in parsed_metadata.items():
            s = Station()
            s.uid = station_code
            s.name = station_metadata['nwsli']
            s.location = sPoint(station_metadata['longitude'],
                                station_metadata['latitude'],
                                0)  # hads always vertically zero

            s.set_property("location_description" , station_metadata['location_text'])
            s.set_property("state"                , station_metadata['state'])
            s.set_property("country"              , "USA")   # @TODO
            s.set_property("vertical_units"       , "ft")
            s.set_property("horizontal_crs"       , "EPSG:4326")
            s.set_property("vertical_crs"         , None)
            s.set_property("hsa"                  , station_metadata['hsa'])
            s.set_property("init_transmit"        , station_metadata['init_transmit'])
            s.set_property("manufacturer"         , station_metadata['manufacturer'])
            s.set_property("owner"                , station_metadata['owner'])
            s.set_property("channel"              , station_metadata['channel'])

            stations.append(s)

            # data

            # possibility no data for this station, or vars filtered all out
            if station_code not in parsed_data:
                continue

            # need to group into distinct time/z value pairs

            # create a keyfunc (creates string of <z>-<timestamp>)
            zandtime = lambda x: str(x[3]) + "-" + str(time.mktime(x[1].timetuple()))

            # annotate data with z values, sort, group by keyfunc (z/time)
            grouped_data = groupby(sorted([(x[0],
                                            x[1],
                                            x[2],
                                            parsed_metadata[station_code]['variables'][x[0]]['base_elevation']) for
            x in parsed_data[station_code]], key=zandtime), zandtime)

            for _, group in grouped_data:

                # group is an iterator, turn it into a list (it will have at least one item)
                groupvals = list(group)

                p = Point()
                p.time = groupvals[0][1]
                p.location = sPoint(station_metadata['longitude'],
                                    station_metadata['latitude'],
                                    groupvals[0][3])

                for val in groupvals:
                    std_var = self.get_variable_info(val[0])
                    if std_var is None:
                        print("Unknown PE Code, ignoring: {} (station: {}).".format(val[0], station_code))
                        continue

                    p.add_member(Member(value=val[2],
                                        standard=std_var[0],
                                        unit=std_var[1],
                                        name=std_var[2],
                                        description=std_var[3]))

                s.add_element(p)

        return StationCollection(elements=stations)

    def _parse_data(self, raw_data, var_filter, time_extents):
        """
        Transforms raw HADS observations into a dict:
            station code -> [(variable, time, value), ...]

        Takes into account the var filter (if set).
        """

        retval = defaultdict(list)
        p = parser()

        begin_time, end_time = time_extents

        for line in raw_data.splitlines():
            if len(line) == 0:
                continue

            fields = line.split("|")[0:-1]
            if var_filter is None or fields[2] in var_filter:
                dt = p.parse(fields[3]).replace(tzinfo=pytz.utc)
                if (begin_time is None or dt >= begin_time) and (end_time is None or dt <= end_time):
                    value = float(fields[4]) if fields[4] != 'NaN' else npNan
                    retval[fields[0]].append((fields[2], dt, value))

        return dict(retval)

    def _parse_metadata(self, metadata):
        """
        Transforms raw HADS metadata into a dictionary (station code -> props)
        """
        retval = {}

        # these are the first keys, afterwards follows a var-len list of variables/props
        # first key always blank so skip it
        field_keys = ['nesdis_id',
                      'nwsli',
                      'location_text',
                      'latitude',
                      'longitude',
                      'hsa',
                      'state',
                      'owner',
                      'manufacturer',
                      'channel',
                      'init_transmit',  # HHMM
                      'trans_interval']  # min

        # repeat in blocks of 7 after field_keys
        var_keys = ['pe_code',
                    'data_interval',  # min
                    'coefficient',
                    'constant',
                    'time_offset',  # min
                    'base_elevation',  # ft
                    'gauge_correction']  # ft

        lines = metadata.splitlines()
        for line in lines:
            if len(line) == 0:
                continue

            raw_fields = line.split("|")

            fields = dict(zip(field_keys, raw_fields[1:len(field_keys)]))

            # how many blocks of var_keys after initial fields
            var_offset = len(field_keys) + 1
            var_blocks = (len(raw_fields) - var_offset) // len(var_keys)     # how many variables
            vars_only  = raw_fields[var_offset:]
            variables  = {}

            for offset in range(var_blocks):
                var_dict = dict(zip(var_keys, vars_only[offset*len(var_keys):(offset+1)*len(var_keys)]))
                variables[var_dict['pe_code']] = var_dict

                var_dict['base_elevation'] = float(var_dict['base_elevation'])
                var_dict['gauge_correction'] = float(var_dict['gauge_correction'])
                del var_dict['pe_code']  # no need to duplicate

            line_val = {'variables' : variables}
            line_val.update(fields)

            # conversions
            def dms_to_dd(dms):
                parts = dms.split(" ")
                sec   = int(parts[1]) * 60 + int(parts[2])
                return float(parts[0]) + (sec / 3600.0)  # negative already in first portion

            line_val['latitude']  = dms_to_dd(line_val['latitude'])
            line_val['longitude'] = dms_to_dd(line_val['longitude'])

            retval[line_val['nesdis_id']] = line_val

        return retval

    @classmethod
    def get_variable_info(cls, hads_var_name):
        """
        Returns a tuple of (mmi name, units, english name, english description) or None.
        """
        if hads_var_name == "UR":
            return ("wind_gust_from_direction", "degrees from N", "Wind Gust from Direction", "Direction from which wind gust is blowing when maximum wind speed is observed.  Meteorological Convention. Wind is motion of air relative to the surface of the earth.")
        elif hads_var_name in ["VJA", "TX"]:
            return ("air_temperature_maximum", "f", "Air Temperature Maximum", "")
        elif hads_var_name in ["VJB", "TN"]:
            return ("air_temperature_minimum", "f", "Air Temperature Minumum", "")
        elif hads_var_name == "PC":  # PC2?
            return ("precipitation_accumulated", "in", "Precipitation Accumulated", "Amount of liquid equivalent precipitation accumulated or totaled for a defined period of time, usually hourly, daily, or annually.")
        elif hads_var_name == "PP":
            return ("precipitation_rate", "in", "Precipitation Rate", "Amount of wet equivalent precipitation per unit time.")
        elif hads_var_name == "US":
            return ("wind_speed", "mph", "Wind Speed", "Magnitude of wind velocity. Wind is motion of air relative to the surface of the earth.")
        elif hads_var_name == "UD":
            return ("wind_from_direction", "degrees_true", "Wind from Direction", "Direction from which wind is blowing.  Meteorological Convention. Wind is motion of air relative to the surface of the earth.")
        elif hads_var_name in ["UP", "UG", "VUP"]:
            return ("wind_gust", "mph", "Wind Gust Speed", "Maximum instantaneous wind speed (usually no more than but not limited to 10 seconds) within a sample averaging interval. Wind is motion of air relative to the surface of the earth.")
        elif hads_var_name in ["TA", "TA2"]:
            return ("air_temperature", "f", "Air Temperature", "Air temperature is the bulk temperature of the air, not the surface (skin) temperature.")
        elif hads_var_name == "MT":
            return ("fuel_temperature", "f", "Fuel Temperature", "")
        elif hads_var_name == "XR":
            return ("relative_humidity", "percent", "Relative Humidity", "")
        elif hads_var_name == "VB":
            return ("battery_voltage", "voltage", "Battery Voltage", "")
        elif hads_var_name == "MM":
            return ("fuel_moisture", "percent", "Fuel Moisture", "")
        elif hads_var_name == "RW":
            return ("solar_radiation", "watt/m^2", "Solar Radiation", "")
        elif hads_var_name == "RS":
            return ("photosynthetically_active_radiation", "watt/m^2", "Photosynthetically Active Radiation", "")
        elif hads_var_name == "TW":     # TW2?
            return ("sea_water_temperature", "f", "Sea Water Temperature", "Sea water temperature is the in situ temperature of the sea water.")
        elif hads_var_name == "WT":
            return ("turbidity", "nephelometric turbidity units", "Turbidity", "")
        elif hads_var_name == "WC":
            return ("sea_water_electrical_conductivity", "micro mhos/cm", "Sea Water Electrical Conductivity", "")
        elif hads_var_name == "WP":
            return ("sea_water_ph_reported_on_total_scale", "std units", "Sea Water PH reported on Total Scale", "the measure of acidity of seawater")
        elif hads_var_name == "WO":
            return ("dissolved_oxygen", "ppm", "Dissolved Oxygen", "")
        elif hads_var_name == "WX":
            return ("dissolved_oxygen_saturation", "percent", "Dissolved Oxygen Saturation", "")
        elif hads_var_name == "TD":
            return ("dew_point_temperature", "f", "Dew Point Temperature", "the temperature at which a parcel of air reaches saturation upon being cooled at constant pressure and specific humidity.")
        elif hads_var_name == "HG":  # HG2?
            return ("stream_gage_height", "ft", "Stream Gage Height", "")
        elif hads_var_name == "HP":
            return ("water_surface_height_above_reference_datum", "ft", "Water Surface Height Above Reference Datum", "means the height of the upper surface of a body of liquid water, such as sea, lake or river, above an arbitrary reference datum.")
        elif hads_var_name == "WS":
            return ("salinity", "ppt", "Salinity", "")
        elif hads_var_name == "HM":
            return ("water_level", "ft", "Water Level", "")
        elif hads_var_name == "PA":
            return ("air_pressure", "hp", "Air Pressure", "")
        elif hads_var_name == "SD":
            return ("snow_depth", "in", "Snow Depth", "")
        elif hads_var_name == "SW":
            return ("snow_water_equivalent", "m", "Snow Water Equivalent", "")
        elif hads_var_name == "TS":
            return ("soil_temperature", "f", "Soil Temperature", "Soil temperature is the bulk temperature of the soil, not the surface (skin) temperature.")

        return None



from __future__ import (absolute_import, division, print_function)
from six import string_types

from pyoos.collectors.collector import Collector
import requests
from pyoos.parsers.waterml import WaterML11ToPaegan


class UsgsRest(Collector):
    def __init__(self, **kwargs):
        super(UsgsRest, self).__init__()
        self.rest_url = 'http://waterservices.usgs.gov/nwis/iv'
        self._state = None

    def set_state(self, state):
        if state is not None:
            if not isinstance(state, string_types):
                raise ValueError("Not a recognized state. \
                                  Must be a str or unicode string")
        self._state = state

    def get_state(self):
        return self._state
    state = property(get_state, set_state)

    def list_variables(self):
        return "Find variables using this website: http://help.waterdata.usgs.gov/codes-and-parameters/parameters"

    def list_features(self):
        return "Find features using this mapping client: http://maps.waterdata.usgs.gov/mapper/"

    def clear(self):
        super(UsgsRest, self).clear()
        self._state = None

    def setup_params(self, **kwargs):
        params = kwargs

        majors = [_f for _f in [self.bbox, self.features, self.state] if _f]
        if len(majors) > 1:
            raise ValueError("""USGS does not allow filtering by more than one 'major' filter.
                                State, BBox, and Features (sites) are all considered 'major' filters.
                                Please set all but one of the filters to None and try again.""")

        if self.start_time is not None:
            params["startDT"] = self.start_time.strftime('%Y-%m-%dT%H:%M')
        if self.end_time is not None:
            params["endDT"] = self.end_time.strftime('%Y-%m-%dT%H:%M')
        if self.bbox is not None:
            params["bBox"] = ",".join([str(x) for x in self.bbox])
        if self.variables is not None and len(self.variables) > 0:
            params["parameterCd"] = ",".join(self.variables)
        if self.features is not None and len(self.features) > 0:
            params["sites"] = ",".join(self.features)
        if self.state is not None:
            params["stateCd"] = self.state

        # Get WaterML 1.1
        params["format"] = "waterml,1.1"

        return params

    def collect(self):
        params = self.setup_params()
        data = requests.get(self.rest_url, params=params).text
        return WaterML11ToPaegan(data).feature

    def raw(self, **kwargs):
        params = self.setup_params()
        return requests.get(self.rest_url, params=params).text



from __future__ import (absolute_import, division, print_function)

import unittest
import pytest

from pyoos.collectors.awc.awc_rest import AwcRest
from paegan.cdm.dsg.collections.station_collection import StationCollection


class AwcRestTest(unittest.TestCase):

    def setUp(self):
        self.c = AwcRest()

    @pytest.mark.xfail
    def test_nwc_stations(self):
        # See https://github.com/ioos/pyoos/issues/63
        stations = self.c.stations
        assert stations[0] == 'AGGH'
        assert stations[-1] == 'ZYTX'

    def test_bbox_filter_raw(self):
        self.c.filter(bbox=(-80, 30, -60, 50))
        response = self.c.raw()
        assert '<response xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XML-Schema-instance" version="1.2" xsi:noNamespaceSchemaLocation="http://aviationweather.gov/adds/schema/metar1_2.xsd">' in response[0]

    def test_bigbbox_filter_paegan(self):
        self.c.filter(bbox=(-80, 30, -60, 50))
        response = self.c.collect()
        assert type(response) == StationCollection

    def test_smallbbox_filter_paegan(self):
        self.c.filter(bbox=(-74, 41, -73, 43))
        response = self.c.collect()
        assert type(response) == StationCollection



from __future__ import (absolute_import, division, print_function)

import unittest
from pyoos.collectors.hads.hads import Hads


class HadsTest(unittest.TestCase):

    def setUp(self):
        self.c = Hads()

        # this is a unit test.  don't hit any urls
        def list_vars(station_codes):
            return set([u'VB', u'DJ', u'DH', u'TW', u'HP', u'UP', u'US', u'PC', u'PA', u'HT', u'UD', u'TA'])
        self.c._list_variables = list_vars

        def state_urls():
            return ["http://amazon.nws.noaa.gov/hads/charts/RI.html"]
        self.c._get_state_urls = state_urls

        def get_stations_for_state(url):
            # all RI stations
            return ['17E3D706', '16C9437A', '1732F0EA', '16C95EDE', '17BC53C2', '17B56004', 'CE4D0268', '1732E39C', '3B0335EE', '3B0211F8', '83779774', 'CD0362DC', 'CD035746', 'DD182264', '16C91306', 'CE2F83BA', '17BC752E', 'CE467750', '1789841A']
        self.c._get_stations_for_state = get_stations_for_state

    def test_set_bbox_property_clears_cached_station_codes(self):
        self.c.station_codes = [1, 2]
        self.c.bbox = (1, 2, 3, 4)

        assert self.c.station_codes is None

    def test_set_features_property_clears_cached_station_codes(self):
        self.c.station_codes = [1, 2]
        self.c.features = ["one", "two"]

        assert self.c.station_codes is None

    def test_list_features(self):
        codes = self.c.list_features()
        assert 'DD182264' in codes

    def test_feature_filter(self):
        features = ['DD182264', '17BC752E', 'CE4D0268']
        self.c.filter(features=features)

        codes = self.c.list_features()

        assert set(codes) == set(features)



from __future__ import (absolute_import, division, print_function)

from owslib.namespaces import Namespaces
from owslib.util import testXMLValue, testXMLAttribute, extract_time
from owslib.util import nspath_eval
from owslib.crs import Crs

from shapely.geometry import box, Point

from pyoos.parsers.ioos.get_observation import IoosGetObservation
from pyoos.parsers.ioos.one.timeseries import TimeSeries
from pyoos.parsers.ioos.one.timeseries_profile import TimeSeriesProfile


def get_namespaces():
    ns = Namespaces()
    return ns.get_namespaces(["om10", "swe101", "swe20", "gml311", "xlink"])
namespaces = get_namespaces()


def nspv(path):
    return nspath_eval(path, namespaces)


class GetObservation(IoosGetObservation):
    def __init__(self, element):
        super(GetObservation, self).__init__(element=element)

        self.ioos_version = "1.0"

        for ob in self._root.findall(nspv("om10:member/om10:Observation")):
            ob_ele = OmObservation(ob)
            self.observations.append(ob_ele)


class OmObservation(object):
    def __init__(self, element):
        self._root = element

        self.description = testXMLValue(self._root.find(nspv("gml311:description")))

        self.begin_position = extract_time(self._root.find(nspv('om10:samplingTime/gml311:TimePeriod/gml311:beginPosition')))

        self.end_position = extract_time(self._root.find(nspv('om10:samplingTime/gml311:TimePeriod/gml311:endPosition')))

        self.procedures = [testXMLAttribute(e, nspv("xlink:href")) for e in self._root.findall(nspv("om10:procedure/om10:Process/gml311:member"))]

        self.observedProperties = [testXMLAttribute(e, nspv("xlink:href")) for e in self._root.findall(nspv("om10:observedProperty/swe101:CompositePhenomenon/swe101:component"))]

        # Can't use a full Xpath expression, so iterate over all metaDataProperties to find the IOOS FeatureType
        self.feature_type = None
        ft = self._root.findall(nspv("om10:featureOfInterest/gml311:FeatureCollection/gml311:metaDataProperty/gml311:GenericMetaData/gml311:name"))
        ft.extend(self._root.findall(nspv("om10:featureOfInterest/gml311:FeatureCollection/gml311:metaDataProperty/gml311:name")))
        ft_def = "http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.6/cf-conventions.html#discrete-sampling-geometries"
        for f in ft:
            if testXMLAttribute(f, "codeSpace") == ft_def:
                self.feature_type = testXMLValue(f)

        # BBOX
        envelope = self._root.find(nspv("om10:featureOfInterest/gml311:FeatureCollection/gml311:boundedBy/gml311:Envelope"))
        self.bbox_srs = Crs(testXMLAttribute(envelope, 'srsName'))
        lower_left_corner = testXMLValue(envelope.find(nspv('gml311:lowerCorner'))).split(" ")
        upper_right_corner = testXMLValue(envelope.find(nspv('gml311:upperCorner'))).split(" ")
        if self.bbox_srs.axisorder == "yx":
            self.bbox = box(float(lower_left_corner[1]), float(lower_left_corner[0]), float(upper_right_corner[1]), float(upper_right_corner[0]))
        else:
            self.bbox = box(float(lower_left_corner[0]), float(lower_left_corner[1]), float(upper_right_corner[0]), float(upper_right_corner[1]))

        # LOCATION
        location = self._root.find(nspv("om10:featureOfInterest/gml311:FeatureCollection/gml311:location"))
        # Should only have one child
        geo = list(location)[-1]
        self.location = {}

        def get_point(element, srs):
            name  = testXMLValue(element.find(nspv("gml311:name")))
            point = testXMLValue(element.find(nspv("gml311:pos"))).split(" ")
            if srs.axisorder == "yx":
                point = Point(float(point[1]), float(point[0]))
            else:
                point = Point(float(point[0]), float(point[1]))
            return name, point

        self.location_srs = Crs(testXMLAttribute(geo, "srsName"))
        if geo.tag == nspv("gml311:Point"):
            n, p = get_point(geo, self.location_srs)
            self.location[n] = p
        elif geo.tag == nspv("gml311:MultiPoint"):
            for point in geo.findall(nspv("gml311:pointMembers/gml311:Point")):
                n, p = get_point(point, self.location_srs)
                self.location[n] = p

        # Now the fields change depending on the FeatureType
        self.results = self._root.find(nspv("om10:result"))

        # TODO: This should be implemented as a Factory
        self.feature = None
        data = self.results.find(nspv("swe20:DataRecord"))
        if data is not None:
            if self.feature_type == 'timeSeries':
                self.feature = TimeSeries(data).feature
            elif self.feature_type == 'timeSeriesProfile':
                self.feature = TimeSeriesProfile(data).feature
            else:
                print("No feature type found.")



from __future__ import (absolute_import, division, print_function)

from pyoos.utils.etree import ElementType, etree
from owslib.namespaces import Namespaces

ns = Namespaces()
SML_NS = ns.get_versioned_namespace('sml', '1.0.1')
SWE_NS = [ns.get_versioned_namespace('swe', '1.0.1')]

class IoosDescribeSensor(object):

    def __new__(cls, element):
        if isinstance(element, ElementType):
            root = element
        else:
            root = etree.fromstring(element)

        sml_str = ".//{{{0}}}identifier/{{{0}}}Term[@definition='http://mmisw.org/ont/ioos/definition/%s']".format(SML_NS)

        if hasattr(root, 'getroot'):
            root = root.getroot()

        # Circular dependencies are bad. consider a reorganization
        # find the the proper type for the DescribeSensor.
        from pyoos.parsers.ioos.one.describe_sensor import (NetworkDS,
                                                            StationDS, SensorDS)
        for ds_type, constructor in [('networkID', NetworkDS), ('stationID', StationDS), ('sensorID', SensorDS)]:
            if root.find(sml_str % ds_type) is not None:
                return super(IoosDescribeSensor, cls).__new__(constructor)

        # NOAA CO-OPS
        sml_str = ".//{{{0}}}identifier/{{{0}}}Term[@definition='urn:ioos:def:identifier:NOAA::networkID']".format(SML_NS)
        if root.find(sml_str) is not None:
            return super(IoosDescribeSensor, cls).__new__(NetworkDS)

        # If we don't find the proper request from the IOOS definitions,
        # try to adapt a generic DescribeSensor request to the dataset.
        from pyoos.parsers.ioos.one.describe_sensor import GenericSensor
        return super(IoosDescribeSensor, cls).__new__(GenericSensor)



from __future__ import (absolute_import, division, print_function)
import six
import inspect

# Makes etree, ElementType, and ParseError availablefrom either `lxml `or `xml`.
try:
    from lxml import etree
    from lxml.etree import ParseError
    ElementType = etree._Element
except ImportError:
    try:
        # Python 2.x/3.x with ElementTree included.
        import xml.etree.ElementTree as etree

        try:
            from xml.etree.ElementTree import ParseError
        except ImportError:
            from xml.parsers.expat import ExpatError as ParseError

        if hasattr(etree, 'Element') and inspect.isclass(etree.Element):
            # Python 3.4, 3.3, 2.7.
            ElementType = etree.Element
        else:
            # Python 2.6.
            ElementType = etree._ElementInterface

    except ImportError:
        try:
            # Python < 2.5 with ElementTree installed.
            import elementtree.ElementTree as etree
            ParseError = StandardError
            ElementType = etree.Element
        except ImportError:
            raise RuntimeError('You need either lxml or ElementTree to use pyoos!')



from __future__ import (absolute_import, division, print_function)

from dateutil import parser
from copy import copy

from owslib.crs import Crs
from owslib.util import nspath_eval
from owslib.namespaces import Namespaces
from shapely.geometry import Point as sPoint

from paegan.cdm.dsg.member import Member
from paegan.cdm.dsg.features.station import Station
from paegan.cdm.dsg.collections.station_collection import StationCollection
from paegan.cdm.dsg.features.base.point import Point

from owslib.swe.common import Time, DataChoice, DataRecord, AbstractSimpleComponent


def get_namespaces():
    ns = Namespaces()
    return ns.get_namespaces(["swe20"])
namespaces = get_namespaces()


def nspv(path):
    return nspath_eval(path, namespaces)


class TimeSeries(object):
    def __init__(self, element):

        record = DataRecord(element)

        # Top level org structure
        stations_field = record.get_by_name("stations")
        stations = {}
        sensors = {}
        for station in stations_field.content.field:
            s = Station()
            s.name  = station.name
            s.uid   = station.content.get_by_name("stationID").content.value

            # Location
            vector  = station.content.get_by_name("platformLocation").content
            srss = vector.referenceFrame.split("&amp;")
            hsrs = None
            try:
                hsrs = Crs(srss[0])
            except ValueError:
                pass

            vsrs = None
            try:
                vsrs = Crs(srss[-1].replace("2=http:", "http:"))
            except ValueError:
                pass

            s.set_property("horizontal_srs", hsrs)
            s.set_property("vertical_srs",   vsrs)
            s.set_property("localFrame",     vector.localFrame)

            lat = vector.get_by_name("latitude").content.value
            lon = vector.get_by_name("longitude").content.value
            z   = vector.get_by_name("height").content.value

            loc = [lon, lat]
            if z:
                loc.append(z)

            s.location = sPoint(*loc)

            # Sensors
            for sensor in station.content.get_by_name("sensors").content.field:
                name        = sensor.name
                uri         = sensor.content.get_by_name("sensorID").content.value
                height      = None
                location_quantity = sensor.content.get_by_name("height").content
                if location_quantity.referenceFrame == "#%s_frame" % s.name:
                    # Uses the station as reference frame
                    if location_quantity.value and z:
                        height      = z + location_quantity.value
                    horizontal_srs  = s.get_property("horizontal_srs")
                    vertical_srs    = s.get_property("vertical_srs")
                else:
                    # Uses its own height
                    if location_quantity.value:
                        height      = location_quantity.value
                    horizontal_srs  = None
                    vertical_srs    = None
                    if hasattr(sensor, 'referenceFrame'):
                        srss            = sensor.referenceFrame.split("&amp;")
                        try:
                            horizontal_srs = Crs(srss[0])
                        except ValueError:
                            pass
                        try:
                            vertical_srs = Crs(srss[-1].replace("2=http:", "http:"))
                        except ValueError:
                            pass

                loc = [s.location.x, s.location.y]
                if height:
                    loc.append(height)

                location            = sPoint(*loc)

                sensors[name] = {   'station'           : s.uid,
                                    'name'              : name,
                                    'uri'               : uri,
                                    'horizontal_srs'    : horizontal_srs,
                                    'vertical_srs'      : vertical_srs,
                                    'location'          : location,
                                    'columns'           : [],  # Array of Members representing the columns
                                    'values'            : []   # Array of Points (the actual data)
                                }

            stations[s.uid] = s

        # Start building the column structure
        data_array = record.get_by_name("observationData").content
        data_record = data_array.elementType.content

        columns = []
        # Data outside of the <field name="sensors"> DataChoice element
        for f in data_record.field:
            columns.append(f)

        # Data inside of DataChoice
        sensor_data = data_record.get_by_name("sensor")
        for sendata in sensor_data.content.item:
            if sendata.content is not None:
                sensors[sendata.name]['columns'] = []
                sensors[sendata.name]['values'] = []
                for f in sendata.content.field:
                    # Create a model Member for each column that will be copied and filled with data from each row
                    sensors[sendata.name]['columns'].append(f)

        decimalSeparator    = data_array.encoding.decimalSeparator
        tokenSeparator      = data_array.encoding.tokenSeparator
        blockSeparator      = data_array.encoding.blockSeparator
        collapseWhiteSpaces = data_array.encoding.collapseWhiteSpaces

        data_values = data_array.values
        self.raw_data = copy(data_values)

        for row in [x for x in data_values.split(blockSeparator) if x != ""]:

            pt = None
            members     = []
            values      = row.split(tokenSeparator)
            sensor_key  = None
            i           = 0

            for x in columns:

                if isinstance(x.content, Time) and x.content.definition == "http://www.opengis.net/def/property/OGC/0/SamplingTime":
                    pt      = Point()
                    pt.time = parser.parse(values[i])

                elif isinstance(x.content, DataChoice):
                    sensor_key = values[i]
                    dc_cols = sensors[sensor_key]['columns']

                    for j, c in enumerate(dc_cols):
                        if isinstance(c.content, AbstractSimpleComponent):
                            m = Member( units=c.content.uom,
                                        name=c.name,
                                        standard=c.content.definition,
                                        value=float(values[i+1]))
                            members.append(m)

                        elif isinstance(c.content, Time) and c.content.definition == "http://www.opengis.net/def/property/OGC/0/SamplingTime":
                            pt      = Point()
                            pt.time = parser.parse(values[i])

                        # For each data column
                        i += 1

                elif isinstance(x.content, AbstractSimpleComponent):
                    m = Member( units=x.content.uom,
                                name=x.name,
                                standard=x.content.definition,
                                value=float(values[i]))
                    members.append(m)

                else:
                    print("WHAT AM I?")

                i += 1

            pt.members = members
            pt.location = stations[sensors[sensor_key]['station']].location
            sensors[sensor_key]['values'].append(pt)

        for k, v in stations.items():
            for sk, sv in sensors.items():
                # Match on station uid
                if sv['station'] == k:
                    v.elements = self._merge_points(v.elements or [], sv['values'])

        if len(stations) > 1:
            self.feature = StationCollection(elements=stations)
        elif len(stations) == 1:
            self.feature = next(iter(stations.values()))
        else:
            print("No stations found!")

    def _merge_points(self, pc1, pc2):
        """
        Merges points based on time/location.

        @TODO: move to paegan, SO SLOW
        """
        res = pc1[:]

        for p in pc2:
            for sp in res:
                if sp.time == p.time and (sp.location is None or (sp.location.equals(p.location))):
                    sp.members.extend(p.members)
                    break
            else:
                res.append(p)

        return res




from __future__ import (absolute_import, division, print_function)
from six import string_types

import unittest
import csv
import io
from datetime import datetime

from owslib.swe.sensor.sml import SensorML

from pyoos.collectors.coops.coops_sos import CoopsSos


class CoopsSosTest(unittest.TestCase):

    def setUp(self):
        self.c = CoopsSos()

    def test_coops_server_id(self):
        assert self.c.server.identification.title == "NOAA.NOS.CO-OPS SOS"
        assert self.c.server.identification.service == 'OGC:SOS'
        assert self.c.server.identification.version == '1.0.0'
        assert self.c.server.identification.abstract == 'NOAA.NOS.CO-OPS Sensor Observation Service (SOS) Server'
        assert self.c.server.identification.keywords == ['Air Temperature', 'Barometric Pressure', 'Conductivity', 'Currents', 'Datum', 'Harmonic Constituents', 'Rain Fall', 'Relative Humidity', 'Salinity', 'Visibility', 'Water Level', 'Water Level Predictions', 'Water Temperature', 'Winds']
        assert self.c.server.identification.fees == 'NONE'
        assert self.c.server.identification.accessconstraints == 'NONE'

    def test_coops_describe_sensor(self):
        self.c.features = ['8454000']
        response = self.c.metadata(output_format='text/xml;subtype="sensorML/1.0.1/profiles/ioos_sos/1.0"')
        assert isinstance(response[0], SensorML)

    def test_raw_coops_get_observation(self):
        self.c.start_time   = datetime.strptime("2012-10-01", "%Y-%m-%d")
        self.c.end_time     = datetime.strptime("2012-10-02", "%Y-%m-%d")
        self.c.features     = ['8454000']
        self.c.variables    = ['http://mmisw.org/ont/cf/parameter/water_surface_height_above_reference_datum']  # noqa

        response = self.c.raw(responseFormat="text/csv").decode()
        assert isinstance(response, string_types)
        """
        station_id,sensor_id,"latitude (degree)","longitude (degree)",date_time,"water_surface_height_above_reference_datum (m)",datum_id,"vertical_position (m)"
        urn:ioos:station:NOAA.NOS.CO-OPS:8454000,urn:ioos:sensor:NOAA.NOS.CO-OPS:8454000:A1,41.8071,-71.4012,2012-10-01T00:00:00Z,1.465,urn:ioos:def:datum:noaa::MLLW,1.064
        """
        data = list(csv.DictReader(io.StringIO(response)))
        assert data[0]['station_id'] == 'urn:ioos:station:NOAA.NOS.CO-OPS:8454000'
        assert data[0]['datum_id'] == "urn:ioos:def:datum:noaa::MLLW"
        assert data[0]['date_time'] == "2012-10-01T00:00:00Z"
        assert data[0]['water_surface_height_above_reference_datum (m)'] == "1.465"
        assert data[0]['vertical_position (m)'] == "1.064"

    def test_raw_coops_get_observation_with_dataType(self):
        self.c.start_time   = datetime.strptime("2012-10-01", "%Y-%m-%d")
        self.c.end_time     = datetime.strptime("2012-10-02", "%Y-%m-%d")
        self.c.features     = ['8454000']
        self.c.variables    = ['http://mmisw.org/ont/cf/parameter/water_surface_height_above_reference_datum']
        self.c.dataType     = "VerifiedHighLow"

        response = self.c.raw(responseFormat="text/csv").decode()
        assert isinstance(response, string_types)
        """
        station_id,sensor_id,"latitude (degree)","longitude (degree)",date_time,"water_surface_height_above_reference_datum (m)",datum_id,"vertical_position (m)"
        urn:ioos:station:NOAA.NOS.CO-OPS:8454000,urn:ioos:sensor:NOAA.NOS.CO-OPS:8454000:W3,41.8071,-71.4012,2012-10-01T01:00:00Z,1.617,urn:ioos:def:datum:noaa::MLLW,1.064
        """
        data = list(csv.DictReader(io.StringIO(response)))
        assert data[0]['station_id'] == 'urn:ioos:station:NOAA.NOS.CO-OPS:8454000'
        assert data[0]['datum_id'] == "urn:ioos:def:datum:noaa::MLLW"
        assert data[0]['date_time'] == "2012-10-01T01:00:00Z"
        assert data[0]['water_surface_height_above_reference_datum (m)'] == "1.617"
        assert data[0]['vertical_position (m)'] == "1.064"

    def test_raw_coops_get_observation_with_datum(self):
        self.c.start_time   = datetime.strptime("2012-10-01", "%Y-%m-%d")
        self.c.end_time     = datetime.strptime("2012-10-02", "%Y-%m-%d")
        self.c.features     = ['8454000']
        self.c.variables    = ['http://mmisw.org/ont/cf/parameter/water_surface_height_above_reference_datum']
        self.c.dataType     = "VerifiedHighLow"
        self.c.datum        = "NAVD"

        response = self.c.raw(responseFormat="text/csv").decode()
        assert isinstance(response, string_types)
        """
        station_id,sensor_id,"latitude (degree)","longitude (degree)",date_time,"water_surface_height_above_reference_datum (m)",datum_id,"vertical_position (m)"
        urn:ioos:station:NOAA.NOS.CO-OPS:8454000,urn:ioos:sensor:NOAA.NOS.CO-OPS:8454000:W3,41.8071,-71.4012,2012-10-01T01:00:00Z,0.863,urn:ogc:def:datum:epsg::5103,1.818
        """
        data = list(csv.DictReader(io.StringIO(response)))
        assert len(data) == 4
        assert data[0]['station_id'] == 'urn:ioos:station:NOAA.NOS.CO-OPS:8454000'
        assert data[0]['datum_id'] == "urn:ogc:def:datum:epsg::5103"
        assert data[0]['date_time'] == "2012-10-01T01:00:00Z"
        assert data[0]['water_surface_height_above_reference_datum (m)'] == "0.863"
        assert data[0]['vertical_position (m)'] == "1.818"



from __future__ import (absolute_import, division, print_function)
from __future__ import with_statement

import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


def extract_version(module='pyoos'):
    version = None
    fdir = os.path.dirname(__file__)
    fnme = os.path.join(fdir, module, '__init__.py')
    with open(fnme) as fd:
        for line in fd:
            if (line.startswith('__version__')):
                _, version = line.split('=')
                # Remove quotation characters.
                version = version.strip()[1:-1]
                break
    return version


def readme():
    with open('README.rst') as f:
        return f.read()

reqs = [line.strip() for line in open('requirements.txt')]


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup(
    name                = 'pyoos',
    version             = extract_version(),
    description         = 'A Python library for collecting Met/Ocean observations',
    long_description    = readme(),
    license             = 'GPLv3',
    author              = 'Kyle Wilcox',
    author_email        = 'kyle@axiomdatascience.com',
    url                 = 'https://github.com/ioos/pyoos.git',
    packages            = find_packages(exclude=['tests.*', 'tests']),
    install_requires    = reqs,
    tests_require       = ['pytest'],
    cmdclass            = {'test': PyTest},
    classifiers         = [
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering',
        ],
    include_package_data = True,
)



from __future__ import (absolute_import, division, print_function)

from dateutil import parser
from collections import OrderedDict, defaultdict
from bisect import bisect_left

from owslib.crs import Crs
from owslib.namespaces import Namespaces
from shapely.geometry import Point as sPoint

from paegan.cdm.dsg.member import Member
from paegan.cdm.dsg.features.station_profile import StationProfile
from paegan.cdm.dsg.features.base.point import Point
from paegan.cdm.dsg.features.base.profile import Profile
from paegan.cdm.dsg.collections.base.profile_collection import ProfileCollection
from paegan.cdm.dsg.collections.station_collection import StationCollection

from owslib.swe.common import Time, DataChoice, DataRecord


def get_namespaces():
    ns = Namespaces()
    return ns.get_namespaces(["swe20"])
namespaces = get_namespaces()


class ProfileCache(object):
    """
    Helper class to accumulate observations and transform them into ProfileCollections representing TimeseriesProfiles.

    Used internally.
    """
    def __init__(self):
        self._cache = defaultdict(OrderedDict)

    def add_obs(self, sensor, t, obs_point, obs_members):
        """
        """
        profile = self._get_profile(sensor, t)
        point   = self._get_point(profile, obs_point)
        point.members.extend(obs_members)

    def get_collections(self):
        return {k[2] : ProfileCollection(elements=list(pd.values())) for k, pd in self._cache.items()}

    def _get_profile(self, sensor, t):
        sens_loc_tuple = (sensor['location']['point'].x, sensor['location']['point'].y, sensor['station'])
        profile_od     = self._cache[sens_loc_tuple]

        if t not in profile_od:
            profile          = Profile()
            profile.location = sPoint(*sens_loc_tuple[0:2])
            profile.time     = t

            # @TODO this is a hack until we can figure out how to assoc stations properly
            profile.station  = sensor['station']

            profile_od[t]    = profile
            return profile

        return profile_od[t]

    def _get_point(self, profile, point):
        """
        Finds the given point in the profile, or adds it in sorted z order.
        """
        cur_points_z = [p.location.z for p in profile.elements]
        try:
            cur_idx = cur_points_z.index(point.z)
            return profile.elements[cur_idx]
        except ValueError:
            new_idx            = bisect_left(cur_points_z, point.z)
            new_point          = Point()
            new_point.location = sPoint(point)
            new_point.time     = profile.time
            profile.elements.insert(new_idx, new_point)
            return new_point


class TimeSeriesProfile(object):
    def __init__(self, element):
        record = DataRecord(element)

        stations_field = record.get_by_name("stations")

        stations = {}
        sensors = {}

        for station in stations_field.content.field:
            s      = StationProfile()
            s.name = station.name
            s.uid  = station.content.get_by_name("stationID").content.value

            # Location
            vector  = station.content.get_by_name("platformLocation").content
            srss = vector.referenceFrame.split("&amp;")
            hsrs = None
            try:
                hsrs = Crs(srss[0])
            except ValueError:
                pass

            vsrs = None
            try:
                vsrs = Crs(srss[-1].replace("2=http:", "http:"))
            except ValueError:
                pass

            s.set_property("horizontal_srs", hsrs)
            s.set_property("vertical_srs",   vsrs)
            s.set_property("localFrame",     vector.localFrame)

            lat = vector.get_by_name("latitude").content.value
            lon = vector.get_by_name("longitude").content.value
            z   = vector.get_by_name("height").content.value

            loc = [lon, lat]
            if z:
                loc.append(z)

            s.location = sPoint(*loc)

            # sensors
            for sensor in station.content.get_by_name("sensors").content.field:
                name           = sensor.name
                uri            = sensor.content.get_by_name("sensorID").content.value

                sensors[name] = {'station'        : s.uid,
                                 'name'           : name,
                                 'uri'            : uri}
                # orientation
                ori_el = sensor.content.get_by_name("sensorOrientation")
                if ori_el:
                    orientation = self._parse_sensor_orientation(ori_el)
                    sensors[name]['sensor_orientation'] = orientation

                # location
                loc_el = sensor.content.get_by_name("sensorLocation")
                if loc_el:
                    location = self._parse_location(loc_el, s.location)
                    sensors[name]['location'] = location

                # profile bins
                profbins_el = sensor.content.get_by_name("profileBins")
                if profbins_el:
                    profile_bins = self._parse_profile_bins(profbins_el)
                    sensors[name]['profile_bins'] = profile_bins

                # OR profile heights
                profheights_el = sensor.content.get_by_name('profileHeights')
                if profheights_el:
                    profile_heights = self._parse_profile_heights(profheights_el)
                    sensors[name]['profile_heights'] = profile_heights

            s.sensors = sensors

            stations[s.uid] = s

        sensor_data = self._parse_sensor_data(record.get_by_name('observationData'), sensors)

        # sensor data is dict of station id -> profile collection
        for station_id, sensor_profile_data in sensor_data.items():
            stations[station_id].elements.extend(sensor_profile_data._elements)

        if len(stations) > 1:
            self.feature = StationCollection(elements=stations)
        elif len(stations) == 1:
            self.feature = next(iter(stations.values()))
        else:
            print("No stations found!")
            self.feature = None

    def _parse_sensor_orientation(self, ori_el):
        # 'srs':Crs(),    # @TODO (OWSLib cannot parse this Crs yet)
        orientation = {}
        for coord in ori_el.content.coordinate:
            orientation[coord.axisID] = {
                'name' : coord.name,
                'value' : coord.value,
                'axis' : coord.axisID,
                'uom' : coord.uom
            }

        return orientation

    def _parse_location(self, loc_el, station_point):
        vector         = loc_el.content

        srss           = vector.referenceFrame.split("&amp;")

        hsrs = None
        try:
            hsrs = Crs(srss[0])
        except ValueError:
            pass

        vsrs = None
        try:
            vsrs = Crs(srss[-1].replace("2=http:", "http:"))
        except ValueError:
            pass

        local_frame    = vector.localFrame
        lat            = vector.get_by_name("latitude").content.value
        lon            = vector.get_by_name("longitude").content.value
        z              = vector.get_by_name("height").content.value

        loc = [lon, lat]
        if z:
            loc.append(z)
        else:
            loc.append(station_point.z)

        location = {'horizontal_srs' : hsrs,
                    'vertical_srs'   : vsrs,
                    'localFrame'     : local_frame,
                    'point'          : sPoint(*loc)}

        return location

    def _parse_profile_bins(self, profbins_el):
        data_array = profbins_el.content
        count      = int(data_array.elementCount[0].text)
        data       = self._parse_data_array(data_array)

        bin_center_quantity = data_array.elementType.content.get_by_name('binCenter')
        bin_center = {'referenceFrame' : bin_center_quantity.content.referenceFrame,
                      'axisID'         : bin_center_quantity.content.axisID,
                      'uom'            : bin_center_quantity.content.uom,
                      'values'         : data[0]}

        bin_edges_quantityrange = data_array.elementType.content.get_by_name('binEdges')
        bin_edges = {'referenceFrame'  : bin_edges_quantityrange.content.referenceFrame,
                     'axisID'          : bin_edges_quantityrange.content.axisID,
                     'uom'             : bin_edges_quantityrange.content.uom,
                     'values'          : data[1]}

        profile_bins = {'bin_center': bin_center,
                        'bin_edges': bin_edges}

        return profile_bins

    def _parse_profile_heights(self, profheights_el):
        data_array = profheights_el.content
        count      = int(data_array.elementCount[0].text)
        data       = self._parse_data_array(data_array)

        height_el  = data_array.elementType.get_by_name('height')
        profile_definition = {'referenceFrame': height_el.content.referenceFrame,
                              'axisID'        : height_el.content.axisID,
                              'uom'           : height_el.content.uom,
                              'values'        : data[0]}

        return profile_definition

    def _parse_data_array(self, data_array):
        """
        Parses a general DataArray.
        """
        decimalSeparator    = data_array.encoding.decimalSeparator
        tokenSeparator      = data_array.encoding.tokenSeparator
        blockSeparator      = data_array.encoding.blockSeparator
        collapseWhiteSpaces = data_array.encoding.collapseWhiteSpaces

        data_values = data_array.values
        lines = [x for x in data_values.split(blockSeparator) if x != '']

        ret_val = []

        for row in lines:
            values = row.split(tokenSeparator)
            ret_val.append([float(v) if ' ' not in v.strip() else [float(vv) for vv in v.split()] for v in values])

        # transpose into columns
        return [list(x) for x in zip(*ret_val)]

    def _parse_sensor_data(self, obs_el, sensor_info):
        """
        Returns ProfileCollection
        """
        data_array = obs_el.content

        # get column defs
        data_record = data_array.elementType.content
        columns = []
        for f in data_record.field:
            columns.append(f)

        # get more information on sensor cols
        sensor_cols = defaultdict(list)
        sensor_vals = defaultdict(list)

        sensor_rec = data_record.get_by_name('sensor')
        for sendata in sensor_rec.content.item:
            if sendata.content is None:
                continue

            for f in sendata.content.field:
                sensor_cols[sendata.name].append(f)

        # @TODO deduplicate
        decimalSeparator    = data_array.encoding.decimalSeparator
        tokenSeparator      = data_array.encoding.tokenSeparator
        blockSeparator      = data_array.encoding.blockSeparator
        collapseWhiteSpaces = data_array.encoding.collapseWhiteSpaces

        data_values = data_array.values
        lines = [x for x in data_values.split(blockSeparator) if x != '']

        # profile cacher!
        profile_cache = ProfileCache()

        for row in lines:
            values = row.split(tokenSeparator)

            skey     = None
            i        = 0
            cur_time = None
            cur_qual = None

            for c in columns:

                if isinstance(c.content, Time) and c.content.definition == "http://www.opengis.net/def/property/OGC/0/SamplingTime":
                    cur_time = parser.parse(values[i])
                    i += 1

                    if len(c.quality):
                        # @TODO: do some quality constraint checks
                        i += len(c.quality)
                        # for qua in c.quality:

                elif isinstance(c.content, DataChoice) and c.name == "sensor":
                    sensor_key = values[i]
                    i += 1

                    sensor_dr = c.content.get_by_name(sensor_key).content
                    sensor_info_ = sensor_info[sensor_key]
                    parsed, nc = self._parse_sensor_record(sensor_dr, sensor_info_, values[i:])

                    # turn these into Points/Members
                    for rec in parsed:
                        # calc a Z value from rec/sensor and build point
                        point, members = self._build_obs_point(sensor_info_, rec)

                        # add to profile
                        profile_cache.add_obs(sensor_info_, cur_time, point, members)

                    i += nc

        return profile_cache.get_collections()

    def _parse_sensor_record(self, sensor_data_rec, sensor_info, rem_values):
        """
        Parses values via sensor data record passed in.
        Returns parsed values AND how many items it consumed out of rem_values.
        """
        val_idx = 0

        # @TODO seems there is only a single field in each of these
        assert len(sensor_data_rec.field) == 1
        sensor_data_array = sensor_data_rec.field[0].content

        # there is probably not going to be a count in the def, it'll be in the data
        count = None
        count_text = sensor_data_array.elementCount.text
        if count_text:
            count = int(count_text.strip())

        if not count:
            count = int(rem_values[val_idx])
            val_idx += 1

        parsed = []

        for recnum in range(count):
            cur  = []

            for f in sensor_data_array.elementType.field:
                cur_val = rem_values[val_idx]
                val_idx += 1

                m = Member(name=f.name,
                           standard=f.content.definition)

                if hasattr(f.content, 'uom'):
                    m['units'] = f.content.uom

                try:
                    m['value'] = float(cur_val)
                except ValueError:
                    m['value'] = cur_val

                if len(f.quality):
                    m['quality'] = []
                    for qual in f.quality:
                        cur_qual = rem_values[val_idx]
                        val_idx += 1

                        # @TODO check this against constraints
                        m['quality'].append(cur_qual)

                cur.append(m)

            parsed.append(cur)

        return parsed, val_idx

    def _build_obs_point(self, sensor_info, obs_recs):
        """
        Pulls bin/profile height info out and calcs a z.

        TODO: currently extremely naive

        Returns a 2-tuple: point, remaining obs_recs
        """
        cur_point = sensor_info['location']['point']

        keys = [m['name'] for m in obs_recs]
        if 'binIndex' in keys:
            zidx      = keys.index('binIndex')
            bin_index = int(obs_recs[zidx]['value'])
            z         = sensor_info['profile_heights']['values'][bin_index]

            point     = sPoint(cur_point.x, cur_point.y, cur_point.z + z)

        elif 'profileIndex' in keys:
            zidx      = keys.index('profileIndex')
            bin_index = int(obs_recs[zidx]['value'])

            # @TODO take into account orientation, may change x/y/z
            # @TODO bin edges?
            z         = sensor_info['profile_bins']['bin_center']['values'][bin_index]

            point     = sPoint(cur_point.x, cur_point.y, cur_point.z + z)

        else:
            raise ValueError("no binIndex or profileIndex in Member: %s", keys)

        # remove z related Member
        obs_recs = obs_recs[:]
        obs_recs.pop(zidx)

        return point, obs_recs
